----------
-- Name: WOOGLED
-- Description: A player and pet cooldown-tracking addon for WoW
-- Author: Woogs - Greymane (US)
-- Version: 2.0.x
-- License: All Rights Reserved
----------


local WOOGLED = WOOGLED;
local Zones = WOOGLED.Zones;
local Libs = WOOGLED.Libs;
local Spells = WOOGLED.Spells;
local Tools = WOOGLED.Tools;
local Units = WOOGLED.Units;
local CooldownFrames = WOOGLED.CooldownFrames;
local OptionsFrame = WOOGLED.OptionsFrame;
local MinimapFrame = WOOGLED.MinimapFrame;

Units.database = {
	player = {},
	pet = {},
};
Units.numUnits = {
	player = 0,
	pet = 0,
};
Units.numPetLinks = 0;
Units.playerOutsideGroupIndex = {};
Units.playerInGroupIndex = {};
Units.spellRecordCooldown = 500; -- shortest interval, in milliseconds, before new spell cooldown data can be recorded; unless a recordCooldown property is found in the Spells database cooldown table
Units.zoningDeleteBuffer = 15000;

local playerOutsideGroupIndex = Units.playerOutsideGroupIndex;
local playerInGroupIndex = Units.playerInGroupIndex;
local database = Units.database;
local spellRecordCooldown = Units.spellRecordCooldown;
local zoningDeleteBuffer = Units.zoningDeleteBuffer;


local ignore = {
	addUnit = {
		player = {
			event = {
				SPELL_AURA_REMOVED = 1, -- Filters out events from units that may not be present in the zone.
				SPELL_AURA_REMOVED_DOSE = 1, -- Filters out events from units that may not be present in the zone.
				SPELL_PERIODIC_DAMAGE = 1, -- Filters out events from units that may not be present in the zone.
				SPELL_PERIODIC_MISSED = 1, -- Filters out events from units that may not be present in the zone.
				SPELL_PERIODIC_HEAL = 1, -- Filters out events from units that may not be present in the zone.
				SPELL_PERIODIC_ENERGIZE = 1, -- Filters out events from units that may not be present in the zone.
				SPELL_PERIODIC_LEECH = 1, -- Filters out events from units that may not be present in the zone.
			},
			spell = {
				[44521] = 'PvP Preparation', -- Fired before unit affiliation can accurately be determined.
			},
		},
		pet = {
			event = {
				SPELL_AURA_REMOVED = 1, -- Filters out events from units that may not be present in the zone.
				SPELL_AURA_REMOVED_DOSE = 1, -- Filters out events from units that may not be present in the zone.
				SPELL_PERIODIC_DAMAGE = 1, -- Filters out events from units that may not be present in the zone.
				SPELL_PERIODIC_MISSED = 1, -- Filters out events from units that may not be present in the zone.
				SPELL_PERIODIC_HEAL = 1, -- Filters out events from units that may not be present in the zone.
				SPELL_PERIODIC_ENERGIZE = 1, -- Filters out events from units that may not be present in the zone.
				SPELL_PERIODIC_LEECH = 1, -- Filters out events from units that may not be present in the zone.
			},
			spell = {
				[44521] = 'PvP Preparation', -- Fired before unit affiliation can accurately be determined.
			},
		},
	},
	updateUnit = {
		player = {
			event = {
				SPELL_AURA_BROKEN = 1, -- event caster is not the spell caster
				SPELL_AURA_BROKEN_SPELL = 1, -- event caster is not the spell caster
				RANGE_DAMAGE = 1,
				RANGE_MISSED = 1,
				SWING_DAMAGE = 1,
				SWING_MISSED = 1,
			},
		},
		pet = {
			event = {
				SPELL_AURA_BROKEN = 1, -- event caster is not the spell caster
				SPELL_AURA_BROKEN_SPELL = 1, -- event caster is not the spell caster
				RANGE_DAMAGE = 1,
				RANGE_MISSED = 1,
				SWING_DAMAGE = 1,
				SWING_MISSED = 1,
			},
		},
	},
};


-- [PUBLIC METHODS] --


function Units.processEvent(pLogTimestamp, pEvent, pCasterGuid, pCasterName, pTargetGuid, pTargetName, pSpellId)
	
	-- Filter out calls with insufficient parameters.
	if ( pCasterGuid == nil or pCasterGuid == '0x0000000000000000' or pCasterName == nil ) then
		return false;
	end
	
	-- Determine unit type. Filter out unwanted unit types.
	local unitType = Tools.extractUnitType(pCasterGuid);
	if ( Units.database[unitType] == nil ) then
		return false;
	end
	
	local dbIndex = pCasterGuid;
	if ( unitType == 'pet' ) then
		dbIndex = Tools.extractPetId(pCasterGuid);
	end
	
	-- Add new units to the database. Filter out unwanted units based on ignore options.
	if ( database[unitType][dbIndex] == nil and ignore.addUnit[unitType].event[pEvent] == nil and ignore.addUnit[unitType].spell[pSpellId] == nil ) then
		Units.addUnit(unitType, dbIndex, pCasterGuid, pCasterName);
	end
	
	-- If the unit is in the database, check spell id and record data if applicable.
	local unitTable = database[unitType][dbIndex];
	if ( unitTable ~= nil ) then
		if ( ignore.updateUnit[unitType].event[pEvent] == nil ) then
			-- Search spell database.
			local cooldownName, cooldownTable, onlyRecordInterval, class, spec, isHook, resetTable = Spells.getSpellInfo(pSpellId, pEvent);
			-- (1) Record class.
			if ( class ~= nil ) then
				Units.recordClass(unitTable, class);
			end
			-- (2) Record spec.
			if ( spec ~= nil ) then
				Units.recordSpec(unitTable, spec);
			end
			-- (3) Record pet link.
			if ( isHook ) then
				if ( pTargetGuid ~= nil and pCasterGuid ~= pTargetGuid ) then --[[and (unitTable.class == nil or Libs.classes[unitTable.class].pets)]]
					Units.recordPetLink(unitTable, pTargetGuid, pTargetName);
				end
			end
			-- (4) Record cooldown.
			if ( cooldownName ~= nil ) then
				if ( (class == nil or unitTable.class == class) and WOOGLED_OPTIONS.ignoreSpells[cooldownName] == nil ) then
					if ( onlyRecordInterval ) then
						Units.recordCooldownInterval(unitTable, pLogTimestamp, cooldownName, cooldownTable);
					else
						Units.recordCooldown(unitTable, pLogTimestamp, cooldownName, cooldownTable);
					end
				end
			end
			-- (5) Reset cooldowns.
			if ( resetTable ) then
				Units.resetCooldowns(unitTable, resetTable);
			end
		end
		--unitTable.lastEventTimestamp = WOOGLED.time.now;
	end
end


function Units.addUnit(pUnitType, pDbIndex, pGuid, pLogName)
	if ( (pUnitType == 'player' and not Units.playerParseCheck(pLogName)) or (pUnitType == 'pet' and not WOOGLED_OPTIONS.parse.pet) ) then
		return false;
	end
	
	database[pUnitType][pDbIndex] = {
		guid = pGuid,
		logName = pLogName, -- used to check group affiliation
		unitType = pUnitType,
		addTimestamp = WOOGLED.time.now, -- used to remove players that were added because of combat log lag, after zoning
		--lastEventTimestamp = WOOGLED.time.now, -- Used for Player Frame
		cooldowns = { -- indexed by unique cooldown names recieved by Spells (unitType_groupName_cooldownName)
			tables = {}, -- link to cooldown tables in Spells database
			intervals = {}, -- shortest intervals between log timestamps
			maxDurations = {}, -- max durations based on interval tracking
			timestamps = {}, -- local timestamps of last cooldown use
			logTimestamps = {}, -- log timestamps of last cooldown use
		},
	};
	if ( pUnitType == 'player' ) then
		database.player[pDbIndex].name, database.player[pDbIndex].server = Tools.chopUpLogName(pLogName);
		--database.player[pDbIndex].inGroupCount = 0; -- number of times UnitIsFriend() returned true
		--database.player[pDbIndex].notInGroupCount = 0; -- number of times UnitIsFriend() returned false/nil
		--database.player[pDbIndex].groupScale = 0; -- used to keep units in the correct group index (+) = inGroup; (-) = notInGroup
		--if ( UnitIsFriend('player', pLogName) ) then
			--playerInGroupIndex[pDbIndex] = true;
			--database.player[pDbIndex].inGroup = true;
		--else
			--playerOutsideGroupIndex[pDbIndex] = true;
			--database.player[pDbIndex].inGroup = false;
		--end
	end
	Units.updateUnitCounter(pUnitType, 1);
	CooldownFrames.onUnitAdd(database[pUnitType][pDbIndex]);
	return true;
end


function Units.deleteRecentlyAddedUnits()
	for unitType, unitTypeTable in pairs ( Units.database ) do
		for guid, unitTable in pairs ( Units.database[unitType] ) do
			if ( (WOOGLED.time.now - unitTable.addTimestamp) < zoningDeleteBuffer ) then
				Units.deleteUnit(unitType, guid);
			end
		end
	end
end


function Units.deleteUnit(pUnitType, pGuid)
	
	--if ( pUnitType == 'player' ) then
	--	if ( database[pGuid].inGroup ) then
	--		playerInGroupIndex[pGuid] = nil;
	--	else
	--		playerOutsideGroupIndex[pGuid] = nil;
	--	end
	--end
	
	if ( database[pUnitType][pGuid].petLinkTable ~= nil ) then
		local linkedUnitTable = database[pUnitType][pGuid].petLinkTable;
		linkedUnitTable.petLinkTable = nil;
		Units.updatePetLinkCounter(-1);
	end
	
	database[pUnitType][pGuid] = nil;
	Units.updateUnitCounter(pUnitType, -1);
end


function Units.deleteUnitDatabase(pUnitType)
	
	--if ( pUnitType == 'player' ) then
	--	Tools.clearTable(playerInGroupIndex);
	--	Tools.clearTable(playerOutsideGroupIndex);
	--end
	
	Tools.clearTable(Units.database[pUnitType]);
	Units.updateUnitCounter(pUnitType);
	Units.updatePetLinkCounter();
end


function Units.deleteAll()

	-- Clear unique player indexes.
	Tools.clearTable(playerInGroupIndex);
	Tools.clearTable(playerOutsideGroupIndex);
	
	-- Clear unit databases; update database test.
	for unitType, unitTypeTable in pairs ( Units.database ) do
		Tools.clearTable(Units.database[unitType]);
		Units.updateUnitCounter(unitType);
	end
	Units.numPetLinks = 0;
	Units.updatePetLinkCounter();
end


function Units.deleteInvalidUnits()
	for guid, unitTable in pairs ( Units.database.player ) do
		if ( not Units.playerParseCheck(unitTable.logName) ) then
			Units.deleteUnit('player', guid);
		end
	end
	if ( not WOOGLED_OPTIONS.parse.pet ) then
		Units.deleteUnitDatabase('pet');
	end
end


function Units.updateUnitCounter(pUnitType, pAdditive)
	if ( pAdditive ~= nil ) then
		Units.numUnits[pUnitType] = Units.numUnits[pUnitType] + pAdditive;
	else
		Units.numUnits[pUnitType] = 0;
	end
	OptionsFrame.groupFrames['database_total_'..pUnitType..'s']:SetText(Units.numUnits[pUnitType]);
end


function Units.updatePetLinkCounter(pAdditive)
	if ( pAdditive ~= nil ) then
		Units.numPetLinks = Units.numPetLinks + pAdditive;
	else
		Units.numPetLinks = 0;
	end
	OptionsFrame.groupFrames['database_totalPetLinks']:SetText(Units.numPetLinks);
end


-- [PRIVATE METHODS] --


function Units.playerParseCheck(pUnitName)
	-- determines if a player should be added to the database based on parse options
	if ( not WOOGLED_OPTIONS.parse.self and pUnitName == WOOGLED.player.name ) then
		--WOOGLED.print(pUnitName..' is you. Self parse options is toggled off');
		return false;
	end
 	local inGroup = UnitIsFriend('player', pUnitName);
	if ( not WOOGLED_OPTIONS.parse.insideRaid ) then
		if ( inGroup and pUnitName ~= WOOGLED.player.name ) then
			--WOOGLED.print(pUnitName..' is a raid member. Raid member parse option is toggled off');
			return false;
		end
	end
	if ( not WOOGLED_OPTIONS.parse.outsideRaid ) then
		if ( not inGroup ) then
			--WOOGLED.print(pUnitName..' is player outside your raid. Everybody else parse option is toggled off');
			return false;
		end
	end
	--WOOGLED.print(pUnitName..' passed parse test. '..WOOGLED_OPTIONS.parse.self..'-'..WOOGLED_OPTIONS.parse.insideRaid..'-'..WOOGLED_OPTIONS.parse.outsideRaid);
	return true;
end


function Units.recordClass(pUnitTable, pClass)
	if ( pUnitTable.class == nil ) then
		-- first class discovery
		pUnitTable.class = pClass;
		CooldownFrames.onDiscoverClass(pUnitTable.guid);
	end
end


function Units.recordSpec(pUnitTable, pSpec)
	if ( pUnitTable.spec ~= nil ) then
		if ( pUnitTable.spec ~= pSpec and Libs.icons.player.spec[pUnitTable.class][pSpec] ~= nil ) then -- Libs needs to be updated so icon table is not used
			-- spec discovered does not match spec already recorded
			pUnitTable.spec = pSpec;
			CooldownFrames.onDiscoverSpec(pUnitTable.guid);
		end
	else
		-- first spec discovery
		pUnitTable.spec = pSpec;
		CooldownFrames.onDiscoverSpec(pUnitTable.guid);
	end
end


function Units.recordCooldown(pUnitTable, pLogTimestamp, pCooldownName, pCooldownTable)
	local cooldowns = pUnitTable.cooldowns;
	if ( cooldowns.tables[pCooldownName] == nil ) then
		-- first time cooldown is being recorde for this unit
		cooldowns.tables[pCooldownName] = pCooldownTable;
		cooldowns.maxDurations[pCooldownName] = pCooldownTable.maxDuration or pCooldownTable.baseDuration;
	end
	
	-- this cooldown has been recorded for this player before
	if ( cooldowns.logTimestamps[pCooldownName] == nil or (pLogTimestamp - cooldowns.logTimestamps[pCooldownName]) > (pCooldownTable.recordCooldown or spellRecordCooldown) ) then
		Units.recordCooldownInterval(pUnitTable, pLogTimestamp, pCooldownName, pCooldownTable);
		cooldowns.timestamps[pCooldownName] = WOOGLED.time.now;
		cooldowns.logTimestamps[pCooldownName] = pLogTimestamp;
		CooldownFrames.onRecordCooldown(pUnitTable.guid, pCooldownName);
	end
end


function Units.recordCooldownInterval(pUnitTable, pLogTimestamp, pCooldownName, pCooldownTable)
	if ( pUnitTable.cooldowns.logTimestamps[pCooldownName] == nil or pCooldownTable.ignoreInterval or pCooldownTable.recordIntervalOnCast ) then
		return;
	end
	local sinceLast = pLogTimestamp - pUnitTable.cooldowns.logTimestamps[pCooldownName];
	if ( pUnitTable.cooldowns.intervals[pCooldownName] == nil or sinceLast < pUnitTable.cooldowns.intervals[pCooldownName] ) then
		pUnitTable.cooldowns.intervals[pCooldownName] = sinceLast;
		if ( pCooldownTable.allDurations ~= nil ) then
			pUnitTable.cooldowns.maxDurations[pCooldownName] = Tools.getNextLowestVal(sinceLast, pCooldownTable.allDurations);
		else
			pUnitTable.cooldowns.maxDurations[pCooldownName] = pCooldownTable.baseDuration;
		end
	end
	pUnitTable.cooldowns.logTimestamps[pCooldownName] = nil;
end


function Units.recordPetLink(pUnitTable, pTargetGuid, pTargetName)
	local targetUnitType = Tools.extractUnitType(pTargetGuid);
	
	local targetIndex = pTargetGuid;
	if ( targetUnitType == 'pet' ) then
		targetIndex = Tools.extractPetId(pTargetGuid);
	end
	
	if ( Units.database[targetUnitType][targetIndex] == nil ) then
		if ( Units.addUnit(targetUnitType, targetIndex, pTargetGuid, pTargetName) ) then
			pUnitTable.petLinkTable = Units.database[targetUnitType][targetIndex];
			Units.database[targetUnitType][targetIndex].petLinkTable = pUnitTable;
			CooldownFrames.onDiscoverPet(pUnitTable.guid);
			Units.updatePetLinkCounter(1);
		end
	else
		-- TODO: Check for unknown class or make sure pet link is valid between the two classes.
		if ( pUnitTable.petLinkTable ~= Units.database[targetUnitType][targetIndex] ) then
			pUnitTable.petLinkTable = Units.database[targetUnitType][targetIndex];
			Units.database[targetUnitType][targetIndex].petLinkTable = pUnitTable;
			CooldownFrames.onDiscoverPet(pUnitTable.guid);
			Units.updatePetLinkCounter(1);
		end
	end
end


function Units.resetCooldowns(pUnitTable, pResetTable)
	for i, cooldownName in pairs ( pResetTable.cooldowns ) do
		if ( pUnitTable.cooldowns.timestamps[cooldownName] ~= nil ) then
			pUnitTable.cooldowns.timestamps[cooldownName] = 0;
			pUnitTable.cooldowns.logTimestamps[cooldownName] = nil;
		end
	end
	CooldownFrames.onCooldownReset(pUnitTable.guid);
end


function Units.getUnitInfo(pGuid)
	local unitType = Tools.extractUnitType(pGuid);
	if ( Units.database[unitType] ~= nil ) then
		if ( unitType == 'pet' ) then
			local petId = Tools.extractPetId(pGuid);
			return unitType, Units.database[unitType][petId];
		else
			return unitType, Units.database[unitType][pGuid];
		end
	end
	return unitType, nil;
end


--[[function Players.updateUnitGroupStatus()
	for guid, unitTable in pairs ( database ) do
		-- check UnitIsFriend(); adjust scale
		if ( UnitIsFriend('player', unitTable.logName) ) then
			-- UnitIsFriend() returned true
			if ( not unitTable.inGroup ) then
				-- unit is in outside group index; reset scale
				unitTable.groupScale = 0;
			end
			unitTable.groupScale = unitTable.groupScale + 1;
			unitTable.inGroupCount = unitTable.inGroupCount + 1;
		else
			-- UnitIsFriend() returned nil/false
			if ( unitTable.inGroup ) then
				-- unit is in group index; reset scale
				unitTable.groupScale = 0;
			end
			unitTable.groupScale = unitTable.groupScale - 1;
			unitTable.notInGroupCount = unitTable.notInGroupCount + 1;
		end
		-- make sure unit is in the correct group index
		if ( unitTable.groupScale > 1 and not unitTable.inGroup ) then
			-- unit is in the wrong index; switch
			playerOutsideGroupIndex[guid] = nil;
			playerInGroupIndex[guid] = true;
			unitTable.inGroup = true;
		elseif ( database[guid].groupScale < -1 and unitTable.inGroup ) then
			-- unit is in the wrong index; switch
			playerInGroupIndex[guid] = nil;
			if ( unitTable.inGroupCount < 3  ) then
				-- UnitIsFriend() returned true less than 3 times; add unit to outside of group index. used for bg/arenas when friendly players leave
				playerOutsideGroupIndex[guid] = true;
			end
			unitTable.inGroup = false;
		end
	end
end]]